@@ -1,11 +1,13 @@
 # Copyright (c) Jupyter Development Team.
 # Distributed under the terms of the Modified BSD License.
+import asyncio
 import datetime
 import json
 import os
 from logging import Logger
-from queue import Queue
+from queue import Empty, Queue
 from threading import Thread
+from time import monotonic
 from typing import Any, Dict, Optional
 
 import websocket
@@ -29,7 +31,7 @@ class GatewayMappingKernelManager(AsyncMappingKernelManager):
     """Kernel manager that supports remote kernels hosted by Jupyter Kernel or Enterprise Gateway."""
 
     # We'll maintain our own set of kernel ids
-    _kernels: Dict[str, "GatewayKernelManager"] = {}
+    _kernels: Dict[str, "GatewayKernelManager"] = {}  # type:ignore[assignment]
 
     @default("kernel_manager_class")
     def _default_kernel_manager_class(self):
@@ -94,7 +96,7 @@ class GatewayMappingKernelManager(AsyncMappingKernelManager):
         model = None
         km = self.get_kernel(str(kernel_id))
         if km:
-            model = km.kernel
+            model = km.kernel  # type:ignore[attr-defined]
         return model
 
     async def list_kernels(self, **kwargs):
@@ -198,9 +200,28 @@ class GatewayKernelSpecManager(KernelSpecManager):
             return "?user=".join([default_endpoint, kernel_user])
         return default_endpoint
 
+    def _replace_path_kernelspec_resources(self, kernel_specs):
+        """Helper method that replaces any gateway base_url with the server's base_url
+        This enables clients to properly route through jupyter_server to a gateway
+        for kernel resources such as logo files
+        """
+        kernelspecs = kernel_specs["kernelspecs"]
+        for kernel_name in kernelspecs:
+            resources = kernelspecs[kernel_name]["resources"]
+            for resource_name in resources:
+                original_path = resources[resource_name]
+                split_eg_base_url = str.rsplit(original_path, sep="/kernelspecs/", maxsplit=1)
+                new_path = url_path_join(self.parent.base_url, "kernelspecs", split_eg_base_url[1])
+                kernel_specs["kernelspecs"][kernel_name]["resources"][resource_name] = new_path
+                if original_path != new_path:
+                    self.log.debug(
+                        f"Replaced original kernel resource path {original_path} with new "
+                        f"path {kernel_specs['kernelspecs'][kernel_name]['resources'][resource_name]}"
+                    )
+        return kernel_specs
+
     def _get_kernelspecs_endpoint_url(self, kernel_name=None):
         """Builds a url for the kernels endpoint
-
         Parameters
         ----------
         kernel_name : kernel name (optional)
@@ -235,6 +256,7 @@ class GatewayKernelSpecManager(KernelSpecManager):
         self.log.debug(f"Request list kernel specs at: {kernel_spec_url}")
         response = await gateway_request(kernel_spec_url, method="GET")
         kernel_specs = json_decode(response.body)
+        kernel_specs = self._replace_path_kernelspec_resources(kernel_specs)
         return kernel_specs
 
     async def get_kernel_spec(self, kernel_name, **kwargs):
@@ -295,24 +317,27 @@ class GatewayKernelSpecManager(KernelSpecManager):
 class GatewaySessionManager(SessionManager):
     kernel_manager = Instance("jupyter_server.gateway.managers.GatewayMappingKernelManager")
 
-    async def kernel_culled(self, kernel_id):
+    async def kernel_culled(self, kernel_id: str) -> bool:
         """Checks if the kernel is still considered alive and returns true if it's not found."""
-        kernel = None
+        km: Optional[GatewayKernelManager] = None
         try:
+            # Since we keep the models up-to-date via client polling, use that state to determine
+            # if this kernel no longer exists on the gateway server rather than perform a redundant
+            # fetch operation - especially since this is called at approximately the same interval.
+            # This has the effect of reducing GET /api/kernels requests against the gateway server
+            # by 50%!
+            # Note that should the redundant polling be consolidated, or replaced with an event-based
+            # notification model, this will need to be revisited.
             km = self.kernel_manager.get_kernel(kernel_id)
-            kernel = await km.refresh_model()
         except Exception:  # Let exceptions here reflect culled kernel
             pass
-        return kernel is None
-
-
-"""KernelManager class to manage a kernel running on a Gateway Server via the REST API"""
+        return km is None
 
 
 class GatewayKernelManager(AsyncKernelManager):
     """Manages a single kernel remotely via a Gateway Server."""
 
-    kernel_id = None
+    kernel_id: Optional[str] = None  # type:ignore[assignment]
     kernel = None
 
     @default("cache_ports")
@@ -324,7 +349,8 @@ class GatewayKernelManager(AsyncKernelManager):
         self.kernels_url = url_path_join(
             GatewayClient.instance().url, GatewayClient.instance().kernels_endpoint
         )
-        self.kernel_url = self.kernel = self.kernel_id = None
+        self.kernel_url: str
+        self.kernel = self.kernel_id = None
         # simulate busy/activity markers:
         self.execution_state = self.last_activity = None
 
@@ -342,13 +368,13 @@ class GatewayKernelManager(AsyncKernelManager):
 
     def client(self, **kwargs):
         """Create a client configured to connect to our kernel"""
-        kw = {}
+        kw: dict = {}
         kw.update(self.get_connection_info(session=True))
         kw.update(
-            dict(
-                connection_file=self.connection_file,
-                parent=self,
-            )
+            {
+                "connection_file": self.connection_file,
+                "parent": self,
+            }
         )
         kw["kernel_id"] = self.kernel_id
 
@@ -418,22 +444,28 @@ class GatewayKernelManager(AsyncKernelManager):
             if os.environ.get("KERNEL_USERNAME") is None and GatewayClient.instance().http_user:
                 os.environ["KERNEL_USERNAME"] = GatewayClient.instance().http_user
 
+            payload_envs = os.environ.copy()
+            payload_envs.update(kwargs.get("env", {}))  # Add any env entries in this request
+
+            # Build the actual env payload, filtering allowed_envs and those starting with 'KERNEL_'
             kernel_env = {
                 k: v
-                for (k, v) in dict(os.environ).items()
-                if k.startswith("KERNEL_") or k in GatewayClient.instance().env_whitelist.split(",")
+                for (k, v) in payload_envs.items()
+                if k.startswith("KERNEL_") or k in GatewayClient.instance().allowed_envs.split(",")
             }
 
-            # Add any env entries in this request
-            kernel_env.update(kwargs.get("env", {}))
-
             # Convey the full path to where this notebook file is located.
             if kwargs.get("cwd") is not None and kernel_env.get("KERNEL_WORKING_DIR") is None:
                 kernel_env["KERNEL_WORKING_DIR"] = kwargs["cwd"]
 
             json_body = json_encode({"name": kernel_name, "env": kernel_env})
 
-            response = await gateway_request(self.kernels_url, method="POST", body=json_body)
+            response = await gateway_request(
+                self.kernels_url,
+                method="POST",
+                headers={"Content-Type": "application/json"},
+                body=json_body,
+            )
             self.kernel = json_decode(response.body)
             self.kernel_id = self.kernel["id"]
             self.kernel_url = url_path_join(self.kernels_url, url_escape(str(self.kernel_id)))
@@ -464,7 +496,12 @@ class GatewayKernelManager(AsyncKernelManager):
             assert self.kernel_url is not None
             kernel_url = self.kernel_url + "/restart"
             self.log.debug("Request restart kernel at: %s", kernel_url)
-            response = await gateway_request(kernel_url, method="POST", body=json_encode({}))
+            response = await gateway_request(
+                kernel_url,
+                method="POST",
+                headers={"Content-Type": "application/json"},
+                body=json_encode({}),
+            )
             self.log.debug("Restart kernel response: %d %s", response.code, response.reason)
 
     async def interrupt_kernel(self):
@@ -473,7 +510,12 @@ class GatewayKernelManager(AsyncKernelManager):
             assert self.kernel_url is not None
             kernel_url = self.kernel_url + "/interrupt"
             self.log.debug("Request interrupt kernel at: %s", kernel_url)
-            response = await gateway_request(kernel_url, method="POST", body=json_encode({}))
+            response = await gateway_request(
+                kernel_url,
+                method="POST",
+                headers={"Content-Type": "application/json"},
+                body=json_encode({}),
+            )
             self.log.debug("Interrupt kernel response: %d %s", response.code, response.reason)
 
     async def is_alive(self):
@@ -496,16 +538,35 @@ KernelManagerABC.register(GatewayKernelManager)
 class ChannelQueue(Queue):
 
     channel_name: Optional[str] = None
+    response_router_finished: bool
 
     def __init__(self, channel_name: str, channel_socket: websocket.WebSocket, log: Logger):
         super().__init__()
         self.channel_name = channel_name
         self.channel_socket = channel_socket
         self.log = log
+        self.response_router_finished = False
+
+    async def _async_get(self, timeout=None):
+        if timeout is None:
+            timeout = float("inf")
+        elif timeout < 0:
+            raise ValueError("'timeout' must be a non-negative number")
+        end_time = monotonic() + timeout
+
+        while True:
+            try:
+                return self.get(block=False)
+            except Empty as e:
+                if self.response_router_finished:
+                    raise RuntimeError("Response router had finished") from e
+                if monotonic() > end_time:
+                    raise
+                await asyncio.sleep(0)
 
     async def get_msg(self, *args: Any, **kwargs: Any) -> dict:
         timeout = kwargs.get("timeout", 1)
-        msg = self.get(timeout=timeout)
+        msg = await self._async_get(timeout=timeout)
         self.log.debug(
             "Received message on channel: {}, msg_id: {}, msg_type: {}".format(
                 self.channel_name, msg["msg_id"], msg["msg_type"] if msg else "null"
@@ -580,19 +641,21 @@ class GatewayKernelClient(AsyncKernelClient):
 
     # flag for whether execute requests should be allowed to call raw_input:
     allow_stdin = False
-    _channels_stopped = False
-    _channel_queues: Optional[dict] = {}
-    _control_channel: Optional[ChannelQueue]
-    _hb_channel: Optional[ChannelQueue]
-    _stdin_channel: Optional[ChannelQueue]
-    _iopub_channel: Optional[ChannelQueue]
-    _shell_channel: Optional[ChannelQueue]
+    _channels_stopped: bool
+    _channel_queues: Optional[Dict[str, ChannelQueue]]
+    _control_channel: Optional[ChannelQueue]  # type:ignore[assignment]
+    _hb_channel: Optional[ChannelQueue]  # type:ignore[assignment]
+    _stdin_channel: Optional[ChannelQueue]  # type:ignore[assignment]
+    _iopub_channel: Optional[ChannelQueue]  # type:ignore[assignment]
+    _shell_channel: Optional[ChannelQueue]  # type:ignore[assignment]
 
-    def __init__(self, **kwargs):
+    def __init__(self, kernel_id, **kwargs):
         super().__init__(**kwargs)
-        self.kernel_id = kwargs["kernel_id"]
+        self.kernel_id = kernel_id
         self.channel_socket: Optional[websocket.WebSocket] = None
         self.response_router: Optional[Thread] = None
+        self._channels_stopped = False
+        self._channel_queues = {}
 
     # --------------------------------------------------------------------------
     # Channel management methods
@@ -625,13 +688,14 @@ class GatewayKernelClient(AsyncKernelClient):
             enable_multithread=True,
             sslopt=ssl_options,
         )
-        self.response_router = Thread(target=self._route_responses)
-        self.response_router.start()
 
         await ensure_async(
             super().start_channels(shell=shell, iopub=iopub, stdin=stdin, hb=hb, control=control)
         )
 
+        self.response_router = Thread(target=self._route_responses)
+        self.response_router.start()
+
     def stop_channels(self):
         """Stops all the running channels for this kernel.
 
@@ -734,6 +798,11 @@ class GatewayKernelClient(AsyncKernelClient):
             if not self._channels_stopped:
                 self.log.warning(f"Unexpected exception encountered ({be})")
 
+        # Notify channel queues that this thread had finished and no more messages are being received
+        assert self._channel_queues is not None
+        for channel_queue in self._channel_queues.values():
+            channel_queue.response_router_finished = True
+
         self.log.debug("Response router thread exiting...")
 
 
